#include "WaveletOpencl.h"

WaveletOpencl::WaveletOpencl(rect image_size, WaveletOpenclParam *param) : Wavelet(image_size, param)
{
  this->param = param;
  this->valid = (this->selectDevice(this->param->dev_type, this->param->dev_id) && this->createResources());
  if(this->param->print_devices) this->printDevicesInfo();
}

WaveletOpencl::~WaveletOpencl()
{
  this->deleteResources();
}

bool WaveletOpencl::isValid()
{
  return this->valid;
}

void WaveletOpencl::deleteKernels()
  {
    if(this->rgba_gray_kernel != NULL) clReleaseKernel(this->rgba_gray_kernel);
    this->rgba_gray_kernel = NULL;

    if(this->gray_rgba_kernel != NULL) clReleaseKernel(this->gray_rgba_kernel);
    this->gray_rgba_kernel = NULL;

    if(this->program != NULL) clReleaseProgram(this->program);
    this->program = NULL;
  }

void WaveletOpencl::deleteQueue()
  {
    if(this->queue != NULL) clReleaseCommandQueue(this->queue);
    this->queue = NULL;

    if(this->context != NULL) clReleaseContext(this->context);
    this->context = NULL;
  }

void WaveletOpencl::deleteResources()
  {
    this->deleteEvents();
    this->deleteKernels();
    this->deleteBuffers();
    this->deleteQueue();
  }

void WaveletOpencl::deleteBuffers()
{
    if(this->d_image_1 != NULL) clReleaseMemObject(this->d_image_1);
    this->d_image_1 = NULL;

    if(this->d_image_2 != NULL) clReleaseMemObject(this->d_image_2);
    this->d_image_2 = NULL;
}

bool WaveletOpencl::selectDevice(cl_device_type dev_type, cl_uint device_num)
  {
    cl_uint platform_count;
    cl_uint device_count;
    cl_int act_device_num = 0;
    if(clGetPlatformIDs(0, NULL, &platform_count) != CL_SUCCESS){this->error_msg << "Error: clGetPlatformIDs\n"; return false;}
    cl_platform_id *platform_ids = new cl_platform_id[platform_count];
    if(clGetPlatformIDs(platform_count, platform_ids, NULL) != CL_SUCCESS){this->error_msg << "Error: clGetPlatformIDs\n"; return false;}
    for(unsigned int i = 0; i < platform_count; i++)
      {
        if(clGetDeviceIDs(platform_ids[i], dev_type, 0, NULL, &device_count) != CL_SUCCESS) continue;
        if (act_device_num + device_count <= device_num)
          {
            act_device_num += device_count;
            continue;
          }
        cl_device_id *device_ids = new cl_device_id[device_num - act_device_num + 1];
        if (clGetDeviceIDs(platform_ids[i], dev_type, device_num - act_device_num + 1, device_ids, NULL) != CL_SUCCESS) continue;
        this->device = device_ids[device_num - act_device_num];
        delete [] device_ids;
        delete [] platform_ids;
        print_device_info(device_num, this->device);
        return true;
      }
    this->error_msg << "Error: No device found\n";
    return false;
  }

void WaveletOpencl::printDevicesInfo()
  {
    cl_uint platform_count;
    cl_uint device_count;
    cl_uint act_device_num = 0;
    if (clGetPlatformIDs(0, NULL, &platform_count) != CL_SUCCESS){ this->error_msg << "Error: clGetPlatformIDs\n"; return; }
    cl_platform_id *platform_ids = new cl_platform_id[platform_count];
    if(clGetPlatformIDs(platform_count, platform_ids, NULL) != CL_SUCCESS){ this->error_msg << "Error: clGetPlatformIDs\n"; return; }
    for (unsigned int i = 0; i < platform_count; i++)
      {
        if(clGetDeviceIDs(platform_ids[i], CL_DEVICE_TYPE_ALL, 0, NULL, &device_count) != CL_SUCCESS) continue;
        if(device_count == 0) continue;
        cl_device_id *device_ids = new cl_device_id[device_count];
        if(clGetDeviceIDs(platform_ids[i], CL_DEVICE_TYPE_ALL, device_count, device_ids, NULL) != CL_SUCCESS) continue;
        for(unsigned int i = 0; i < device_count; i++)
          {
            print_device_info(act_device_num, device_ids[i]);
            act_device_num++;
          }
        delete [] device_ids;
      }
    delete [] platform_ids;
  }

bool WaveletOpencl::createKernels()
  {
    cl_int err_num;
    const char *src_data = read_file("../kernels/gpu.cl");
    if(src_data == NULL){this->error_msg << "Error: Read kernel data\n"; return false;}
    this->program = clCreateProgramWithSource(this->context, 1, &src_data, NULL, &err_num);
    if(err_num != CL_SUCCESS){this->error_msg << "Error: clCreateProgramWithSource\n"; return false;}
    if((err_num = clBuildProgram(this->program, 1, &this->device, this->param->createBuildParam().c_str(), NULL, NULL)) != CL_SUCCESS)
      {
        this->error_msg << "Error: clBuildProgram\n";
        size_t program_size;
        clGetProgramBuildInfo(this->program, this->device, CL_PROGRAM_BUILD_LOG, NULL, NULL, &program_size);
        char *data = new char[program_size];
        clGetProgramBuildInfo(this->program, this->device, CL_PROGRAM_BUILD_LOG, program_size, data, NULL);
        this->error_msg << "Build log:\n" << data << "\n";
        return false;
      } 
    this->rgba_gray_kernel = clCreateKernel(this->program, "conv_rgba_gray", &err_num);
    if(err_num != CL_SUCCESS){this->error_msg << "Error: clCreateKernel conv_rgba_gray\n"; return false;}
    this->gray_rgba_kernel = clCreateKernel(this->program, "conv_gray_rgba", &err_num);
    if(err_num != CL_SUCCESS){this->error_msg << "Error: clCreateKernel conv_gray_rgba\n"; return false;}
    return true;
  }

bool WaveletOpencl::createQueue()
  {
    cl_int err_num;
    this->context = clCreateContext(NULL, 1, &this->device, NULL, NULL, &err_num);
    if(err_num != CL_SUCCESS){this->error_msg << "Error: clCreateContext\n"; return false;}
    this->queue = clCreateCommandQueue(this->context, this->device, CL_QUEUE_PROFILING_ENABLE, &err_num);
    if(err_num != CL_SUCCESS){this->error_msg << "Error: clCreateCommandQueue\n"; return false;}
    return true;
  }

bool WaveletOpencl::createResources()
{
  if(!this->image_size.isValid() || 
     !this->createQueue() ||
     !this->createKernels() ||
     !this->createBuffers() ||
     !this->createEvents())
    {
      this->deleteResources();
      return false;
    }
  return true;
}

bool WaveletOpencl::createBuffers()
{
  cl_int err_num;
  if((d_image_1 = clCreateBuffer(context, CL_MEM_READ_WRITE, image_size.getSize() * sizeof(cl_uchar4), NULL, &err_num)) == NULL)
    {this->error_msg << "Error: clCreateBuffer d_image_1\n"; return false;}
  if((d_image_2 = clCreateBuffer(context, CL_MEM_READ_WRITE, image_size.getSize() * sizeof(cl_uchar4), NULL, &err_num)) == NULL)
    {this->error_msg << "Error: clCreateBuffer d_image_2\n"; return false;}
  return true;
}

bool WaveletOpencl::createEvents()
{
  this->ev_pre.resize(2);
  this->ev_post.resize(2);
  return true;
}

void WaveletOpencl::deleteEvents()
{
  this->ev_pre.clear();
  this->ev_post.clear();
}

bool WaveletOpencl::getFramePre(uchar4 *bgra_frame)
	{
		if(!this->isValid())
			{
        this->error_msg << "Error: Detector isnt in valid state.\n";
				return false;
			}
    
    for(size_t i = 0; i < this->ev_pre.size(); i++)
      {
        this->ev_pre[i] = clCreateUserEvent(this->context, NULL);
      }
    for(size_t i = 0; i < this->ev_proc.size(); i++)
      {
        this->ev_proc[i] = clCreateUserEvent(this->context, NULL);
      }
    for(size_t i = 0; i < this->ev_post.size(); i++)
      {
        this->ev_post[i] = clCreateUserEvent(this->context, NULL);
      }

    if((clEnqueueWriteBuffer(this->queue, this->d_image_1, CL_FALSE, 0, this->image_size.getSize() * sizeof(cl_uchar4), (void *)bgra_frame, 0, NULL, &(this->ev_pre[0])) != CL_SUCCESS) ||
       (!this->convRgbaGray(this->d_image_2, this->d_image_1, (cl_uint)this->image_size.w, (cl_uint)this->image_size.h, &(this->ev_pre[1])))) return false;

    return true;
	}

bool WaveletOpencl::getFramePost(uchar4 *bgra_frame, cl_mem out_image, cl_mem in_image, det_output *out)
	{
		if(!this->isValid())
			{
        this->error_msg << "Error: Detector isnt in valid state.\n";
				return false;
			}


    if((!this->convGrayRgba(out_image, in_image, (cl_uint)this->image_size.w, (cl_uint)this->image_size.h, &(this->ev_post[0]))) ||
       (clEnqueueReadBuffer(this->queue, out_image, CL_FALSE, 0, this->image_size.getSize() * sizeof(cl_uchar4), (void *)bgra_frame, 0, NULL, &(this->ev_post[1])) != CL_SUCCESS)) return false;
    clFinish(this->queue);

    if(out != NULL)
      {
        for(size_t i = 0; i < this->ev_pre.size(); i++)
          {
            out->pre.push_back(get_time(this->ev_pre[i]));
          }
        for(size_t i = 0; i < this->ev_proc.size(); i++)
          {
            out->proc.push_back(get_time(this->ev_proc[i]));
          }
        for(size_t i = 0; i < this->ev_post.size(); i++)
          {
            out->post.push_back(get_time(this->ev_post[i]));
          }
      }

    for(size_t i = 0; i < this->ev_pre.size(); i++)
      {
        clReleaseEvent(this->ev_pre[i]);
      }
    for(size_t i = 0; i < this->ev_proc.size(); i++)
      {
        clReleaseEvent(this->ev_proc[i]);
      }
    for(size_t i = 0; i < this->ev_post.size(); i++)
      {
        clReleaseEvent(this->ev_post[i]);
      }

    return true;
	}

bool WaveletOpencl::convRgbaGray(cl_mem out_image, cl_mem in_image, cl_uint width, cl_uint height, cl_event *proc_event)
  {
    cl_int err_msg;
    clSetKernelArg(this->rgba_gray_kernel, 0, sizeof(cl_mem), &out_image);
    clSetKernelArg(this->rgba_gray_kernel, 1, sizeof(cl_mem), &in_image);
    clSetKernelArg(this->rgba_gray_kernel, 2, sizeof(cl_uint), &width);
    clSetKernelArg(this->rgba_gray_kernel, 3, sizeof(cl_uint), &height);
    
    size_t local[2] = {32, 8};
    size_t global[2] = {ceil_align(width/4, local[0]), ceil_align(height, local[1])};
    if((err_msg = clEnqueueNDRangeKernel(this->queue, this->rgba_gray_kernel, 2, NULL, global, local, 0, NULL, proc_event)) != CL_SUCCESS){error_msg << "Error: clEnqueueNDRangeKernel rgba_gray_kernel " << err_msg << "\n"; return false;};
    return true;
  }

bool WaveletOpencl::convGrayRgba(cl_mem out_image, cl_mem in_image, cl_uint width, cl_uint height, cl_event *proc_event)
  {
    cl_int err_msg;
    cl_uint depth = this->depth;
    clSetKernelArg(this->gray_rgba_kernel, 0, sizeof(cl_mem), &out_image);
    clSetKernelArg(this->gray_rgba_kernel, 1, sizeof(cl_mem), &in_image);
    clSetKernelArg(this->gray_rgba_kernel, 2, sizeof(cl_uint), &width);
    clSetKernelArg(this->gray_rgba_kernel, 3, sizeof(cl_uint), &height);
    clSetKernelArg(this->gray_rgba_kernel, 4, sizeof(cl_uint), &depth);
    
    size_t local[2] = {32, 8};
    size_t global[2] = {ceil_align(width, local[0]), ceil_align(height, local[1])};
    if((err_msg = clEnqueueNDRangeKernel(this->queue, this->gray_rgba_kernel, 2, NULL, global, local, 0, NULL, proc_event)) != CL_SUCCESS){error_msg << "Error: clEnqueueNDRangeKernel gray_rgba_kernel " << err_msg << "\n"; return false;};
    return true;
  }
